<?php
/*
 * Copyright 2007-2017 Charles du Jeu - Abstrium SAS <team (at) pyd.io>
 * This file is part of Pydio.
 *
 * Pydio is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Pydio is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Pydio.  If not, see <http://www.gnu.org/licenses/>.
 *
 * The latest code can be found at <https://pydio.com>.
 *
 */
namespace Pydio\Access\Driver\StreamProvider\Gateway;

use Psr\Http\Message\ResponseInterface;
use Psr\Http\Message\ServerRequestInterface;
use Pydio\Access\Core\Exception\FileNotFoundException;
use Pydio\Access\Core\Model\Node;
use Pydio\Access\Core\Model\NodesList;
use Pydio\Access\Core\Model\UserSelection;
use Pydio\Access\Core\RecycleBinManager;
use Pydio\Access\Driver\StreamProvider\S3\S3AccessDriver;
use Pydio\Core\Http\Client\MicroApi;
use Pydio\Core\Http\Response\SerializableResponseStream;
use Pydio\Core\Model\ContextInterface;
use Pydio\Core\Services\LocaleService;
use Pydio\Core\Services\SessionService;
use Pydio\Core\Utils\Vars\InputFilter;
use Pydio\Core\Utils\Vars\StatHelper;
use Swagger\Client\ApiException;
use Swagger\Client\Model\RestGetBulkMetaRequest;
use Zend\Diactoros\Response\JsonResponse;

defined('PYDIO_EXEC') or die( 'Access not allowed');

/**
 * Class GatewayAccessDriver
 * @package Pydio\Access\Driver\StreamProvider\Gateway
 */
class GatewayAccessDriver extends S3AccessDriver
{

    public function lsAction(ServerRequestInterface &$request, ResponseInterface &$response){

        /** @var ContextInterface $ctx */
        $ctx = $request->getAttribute("ctx");
        $action = $request->getAttribute("action");
        $httpVars = $request->getParsedBody();

        $selection = UserSelection::fromContext($ctx, $httpVars);
        if (!$selection->isEmpty()) {
            $this->filterUserSelectionToHidden($ctx, $selection->getFiles());
            RecycleBinManager::filterActions($action, $selection, $httpVars);
        }

        $nodesList = new NodesList();
        $dir = trim(InputFilter::sanitize($httpVars["dir"], InputFilter::SANITIZE_DIRNAME), "/") OR "";

        $metaData = [];
        $dirNode = $selection->nodeForPath("/" . $dir);
        if (RecycleBinManager::recycleEnabled() && $dir == "") {
            $dirNode->mergeMetadata(["repo_has_recycle" => RecycleBinManager::getRelativeRecycle()]);
        }
        try{
            $dirNode->setLeaf(false);
            $dirNode->loadNodeInfo(false, true, ($lsOptions["l"]?"all":"minimal"));
        } catch(ApiException $e){
            if($e->getCode() == 404) {
                throw new FileNotFoundException($dirNode->getPath());
            } else {
                throw $e;
            }
        }
        $nodesList->setParentNode($dirNode);

        $api = MicroApi::GetMetaServiceApi();
        $bulkRequest = new RestGetBulkMetaRequest();

        if(!$selection->isEmpty()){

            $selPath = trim($dir, "/") . "/" . $selection->getUniqueFile();
            $bulkRequest->setNodePaths([$ctx->getRepository()->getSlug() . "/" . $selPath]);

        } else {

            $bulkRequest->setNodePaths([$ctx->getRepository()->getSlug() . "/" . ltrim($dir . "/", "/")."*"]);

        }

        try{
            $result = $api->getBulk($bulkRequest);
        } catch(ApiException $e){
            if ($e->getCode() == 404) {
                throw new FileNotFoundException($dirNode->getPath());
            } else {
                throw $e;
            }
        }
        if ($result->getNodes() != null ){
            foreach($result->getNodes() as $idmNode){
                $node = Node::fromApiNode($ctx, $idmNode);
                if (StatHelper::isHidden($node->getLabel())){
                    continue;
                }
                $node->loadNodeInfo();
                $nodesList->addBranch($node);
            }
        }

        $body = new SerializableResponseStream();
        $body->addChunk($nodesList);
        $response = $response->withBody($body);

    }

    /**
     * Update node metadata with core FS metadata.
     * @param \Pydio\Access\Core\Model\Node $ajxpNode
     * @param bool $parentNode
     * @param bool $details
     * @return void
     */
    public function loadNodeInfo(&$ajxpNode, $parentNode = false, $details = false){
        $nodeName = basename($ajxpNode->getPath());
        $metaData = $ajxpNode->metadata;
        $isLeaf = $ajxpNode->isLeaf();

        $mimeData = StatHelper::getMimeInfo($ajxpNode, !$isLeaf);
        $metaData["mimestring_id"] = $mimeData[0];
        $metaData["icon"] = $mimeData[1];
        if(!empty($mimeData[2])){
            $metaData["fonticon"] = $mimeData[2];
        }
        if (!$isLeaf) {
            $metaData["ajxp_mime"] = "ajxp_folder";
        }
        if (StatHelper::isBrowsableArchive($nodeName)) {
            $metaData["ajxp_mime"] = "ajxp_browsable_archive";
        }
        $ajxpNode->mergeMetadata($metaData);
    }


    /**
     * Create a Presigned Url
     * @param ServerRequestInterface $request
     * @param ResponseInterface $response
     */
    public function getPresignedUrl(ServerRequestInterface $request, ResponseInterface &$response) {

        $ctx = $request->getAttribute('ctx');
        $userSelection = UserSelection::fromContext($ctx, $request->getParsedBody());
        $node = $userSelection->getUniqueNode();
        $client = $this->getS3Service($ctx);

        $token = "";
        if (SessionService::has("DEX-AUTH-TOKEN")) {
            $tokens = SessionService::fetch("DEX-AUTH-TOKEN");
            $token = $tokens["id_token"];
        }

        $cmdName = 'GetObject';
        if ($request->getParsedBody()['cmd'] === 'PUT') {
            $cmdName = 'PutObject';
        }
        $nodePath = $node->getRepository()->getSlug() . "/" . ltrim($node->getPath(), "/");
        $cmd = $client->getCommand($cmdName, [
            'Bucket' => 'pydio',
            'Key' => $nodePath,
        ]);
        $req = $client->createPresignedRequest($cmd, '+20 minutes');


        $response = new JsonResponse([
            'path'     => $nodePath,
            'signedUrl'=> (string)$req->getUri(),
            'validity' => '20 minutes',
            'jwt'      => $token,
        ]);
    }

}
