<?php
/*
 * Copyright 2007-2017 Charles du Jeu - Abstrium SAS <team (at) pyd.io>
 * This file is part of Pydio.
 *
 * Pydio is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Pydio is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Pydio.  If not, see <http://www.gnu.org/licenses/>.
 *
 * The latest code can be found at <https://pydio.com>.
 */
namespace Pydio\Access\Meta\Watch;

use Pydio\Access\Core\AbstractAccessDriver;
use Pydio\Access\Core\Model\Node;
use Pydio\Access\Core\Model\UserSelection;
use Pydio\Access\Meta\Core\AbstractMetaSource;
use Pydio\Access\Metastore\Core\IMetaStoreProvider;
use Pydio\Core\Exception\PydioException;
use Pydio\Core\Http\Client\MicroApi;
use Pydio\Core\Model\ContextInterface;
use Pydio\Core\PluginFramework\PluginsService;
use Pydio\Notification\Core\ActivityCenter;

defined('PYDIO_EXEC') or die( 'Access not allowed');

/**
 * Keep an eye on a folder to be alerted when something changes inside it
 * @package AjaXplorer_Plugins
 * @subpackage Meta
 */
class WatchRegister extends AbstractMetaSource
{
    public static $META_WATCH_CHANGE = "META_WATCH_CHANGE";
    public static $META_WATCH_READ = "META_WATCH_READ";
    public static $META_WATCH_BOTH = "META_WATCH_BOTH";
    public static $META_WATCH_NAMESPACE = "META_WATCH";

    public static $META_WATCH_USERS_READ = "META_WATCH_USERS_READ";
    public static $META_WATCH_USERS_CHANGE = "META_WATCH_USERS_CHANGE";
    public static $META_WATCH_USERS_NAMESPACE = "META_WATCH_USERS";

    /**
     * @var IMetaStoreProvider
     */
    protected $metaStore;

    /**
     * @param ContextInterface $ctx
     * @param AbstractAccessDriver $accessDriver
     * @throws PydioException
     */
    public function initMeta(ContextInterface $ctx, AbstractAccessDriver $accessDriver)
    {
        parent::initMeta($ctx, $accessDriver);
        $store = PluginsService::getInstance($ctx)->getUniqueActivePluginForType("metastore");
        if ($store === false) {
            throw new PydioException("The 'meta.watch' plugin requires at least one active 'metastore' plugin");
        }
        $this->metaStore = $store;
        $this->metaStore->initMeta($ctx, $accessDriver);
    }

    /**
     * @param \Psr\Http\Message\ServerRequestInterface $requestInterface
     * @param \Psr\Http\Message\ResponseInterface $responseInterface
     */
    public function switchActions(\Psr\Http\Message\ServerRequestInterface $requestInterface, \Psr\Http\Message\ResponseInterface &$responseInterface)
    {
        $actionName = $requestInterface->getAttribute("action");
        $httpVars = $requestInterface->getParsedBody();
        if($actionName !== "toggle_watch") return;
        /** @var ContextInterface $ctx */
        $ctx = $requestInterface->getAttribute("ctx");

        $us = UserSelection::fromContext($ctx, $httpVars);
        $node = $us->getUniqueNode();
        $node->loadNodeInfo();
        $cmd = $httpVars["watch_action"];

        $nodeUuid = $node->getUuid();
        if(empty($nodeUuid)) {
            throw new PydioException("Cannot find node Uuid");
        }
        $userId = $ctx->hasUser() ? $ctx->getUser()->getId() : "shared";

        $acApi = MicroApi::GetActivityServiceApi();
        $events = [];
        if ($cmd === "watch_change") {
            $events = [ActivityCenter::SUBSCRIBE_CHANGE];
        } else if ($cmd === "watch_read") {
            $events = [ActivityCenter::SUBSCRIBE_READ];
        } else if($cmd === "watch_both") {
            $events = [ActivityCenter::SUBSCRIBE_CHANGE, ActivityCenter::SUBSCRIBE_READ];
        } else if ($cmd == "watch_stop") {
            $events = [];
        }
        ActivityCenter::UserSubscribeToNode($userId, $nodeUuid, $events);

        $node->metadata = array();
        $node->loadNodeInfo(true, false, "all");
        $this->enrichNode($node);
        $nodesDiff = new \Pydio\Access\Core\Model\NodesDiff();
        $nodesDiff->update($node);
        $x = new \Pydio\Core\Http\Response\SerializableResponseStream();
        $x->addChunk($nodesDiff);
        $responseInterface = $responseInterface->withBody($x);

    }


    /**
     * @param Node $node
     */
    public function enrichNode(&$node)
    {
        if(!$node->getContext()->hasUser()) {
            return;
        }
        $uuid = $node->getUuid();
        if(empty($uuid)){
            return;
        }
        $events = [];
        $userId = $node->getContext()->getUser()->getId();
        if($node->metadataLoaded) {
            $subMeta = $node->user_subscriptions;
            if(!empty($subMeta)) {
                $events = explode(",", $subMeta);
            }
        } else {
            // Warning, will trigger a query to the REST API
            $events = ActivityCenter::UserIsSubscribedToNode($userId, $uuid);
        }

        if (count($events)){
            $read = in_array(ActivityCenter::SUBSCRIBE_READ, $events);
            $change = in_array(ActivityCenter::SUBSCRIBE_CHANGE, $events);
            if ($read && $change) {
                $value = self::$META_WATCH_BOTH;
            } else if ($change) {
                $value = self::$META_WATCH_CHANGE;
            } else if ($read) {
                $value = self::$META_WATCH_READ;
            }
            $node->mergeMetadata(array(
                "meta_watched" => $value,
                "overlay_class" => "icon-eye-open"
            ), true);
        }
    }


}
